%{
#include <stdio.h>
#include "itemlist.h"
#include "local.h"
#include "gramactions.h"
   
int   InitRepeat(void);
int   new_state(void);
void  set_module_name (char*); 
int   native = 0;
char *mysprintf(char *format, ...); 
char *numafter (char*,int);
char *numbefore (int,char*);
int line_number;
int absorb;
int parenth_exp;
int parenth_list;

void source (char *text);
void InitAll (void);
int yylex (void);
int yyerror (char *msg);

%}

%union {
   char         *str;
   int           intval;
   local_data_t  param;
   item_list_t   item_list;
   void         *empty;
}

%token MODULE NATIVE IDENTIFIER END 
%token COOPERATE HALT WHILE_TOK REPEAT
%token RUN DO JOIN LOCAL GET_VALUE
%token AWAIT IF THEN ELSE RETURN
%token LINK UNLINK FINALIZE C_STAT
%token PARENTH_EXP PARENTH_LIST

%type <intval> WHILE_TOK RUN UNLINK LINK
%type <intval> COOPERATE HALT REPEAT IF
%type <intval> THEN ELSE AWAIT GET_VALUE JOIN RETURN

%type <str> IDENTIFIER C_STAT PARENTH_EXP
%type <str> num.identifier cooperate halt empty_stat
%type <str> then.branch else.branch while repeat
%type <str> run await if Cstat call join
%type <str> inst atomic.or.non.atomic atomic non.atomic
%type <str> module extern.module native.module
%type <str> return get_value definition
%type <str> link unlink finalizer final

%type <param> parameter
%type <item_list> parameter.list local.declarations
%type <item_list> identifier.list PARENTH_LIST

%type <empty> Cstat_interm call_interm if_interm1 if_interm2
%type <empty> repeat_interm while_interm final_interm1 final_interm2

/************** THE GRAMMAR RULES *************/
%%
initial: {InitAll();} root ;

root: program | /*Empty*/ ;

program:
    definition         {fprintf (stdout,"%s",numafter ($1,line_number));}
  | program definition {fprintf (stdout,"%s",numafter ($2,line_number));}
  |  error END MODULE { absorb = 1; native = 0; fprintf (stdout, "\n/* syntax error */\n"); }
  ;

definition : module | extern.module | native.module ;

/********* modules ***************/
module:       
   module.name '(' parameter.list ')' local.declarations inst finalizer END MODULE
   {$$ = Module($3,$7,$5,$6); absorb = 1; native = 0;}
   ;

module.name: IDENTIFIER {set_module_name ($1);} ;

extern.module:
   identifier.list ';' {$$ = ExternModule($1); absorb = 1;} ;

native.module: NATIVE {native = 1;} module {$$ = $3;} ;

/********* parameters ***************/
parameter.list:
     /*Empty*/
     {$$ = _make_item_list ();}
   | parameter
     {item_list_t l = _make_item_list (); _add_item_list (l, $1); $$ = l;}
   | parameter.list ',' parameter
     {_add_item_list ($1,$3); $$ = $1;}
   ;

num.identifier:
   IDENTIFIER {$$ = (void*)numbefore (line_number,$1);}

parameter:
     num.identifier IDENTIFIER
     {$$ = make_local_data ($1,$2);}
   | num.identifier '*' IDENTIFIER
     {$$ = make_local_data (mysprintf ("%s*",$1),$3);}
   | num.identifier '*' '*' IDENTIFIER
     {$$ = make_local_data (mysprintf ("%s**",$1),$4);}
   ;

identifier.list:
     IDENTIFIER
     {item_list_t l = _make_item_list (); _add_item_list (l, $1); $$ = l;}
   | identifier.list ',' IDENTIFIER
     {_add_item_list ($1,$3); $$ = $1;}
   ;

/********* locals ***************/
local.declarations:
     /*Empty*/                {$$ = NULL;}
   | LOCAL parameter.list ';' {$$ = $2;}
   ;

/********* finalize ***************/
finalizer:
     /*Empty*/   {$$ = NULL;}
   | FINALIZE final {$$ = $2;}
   ;

final_interm1: {parenth_exp = 1; $$ = (void*)line_number;} ;
final_interm2: {$$ = (void*)line_number;} ;

final:
   IDENTIFIER final_interm1 PARENTH_EXP {parenth_exp = 0;} ';'
   {$$ = numbefore((int)$2,Final_Call($1,$3));}
|
   final_interm2 C_STAT {$$ = numbefore ((int)$1,$2);} ;


/********* insts ***************/
inst:
     atomic.or.non.atomic
   | inst atomic.or.non.atomic {$$ = Seq ($1,$2);}
   ;

atomic.or.non.atomic: atomic | non.atomic ;

/***********************************************************/
atomic: Cstat | call | empty_stat ;

Cstat_interm: {$$ = (void*)line_number;} ;

Cstat:
   Cstat_interm C_STAT {$$ = numbefore ((int)$1,Cstat ($2));} ;

call_interm: {parenth_exp=1;$$ = (void*)line_number;} ;

call:
   IDENTIFIER call_interm PARENTH_EXP {parenth_exp=0;} ';'
   {$$ = numbefore((int)$2,Call($1,$3));}
   ;

empty_stat: ';' {$$ = Nothing ();}

/***********************************************************/
non.atomic:
     cooperate
   | halt
   | return
   | await
   | join
   | get_value
   | if
   | while
   | repeat
   | run
   | unlink
   | link
   | error ';' {$$ = "/********** error ***********/";}
   ;

cooperate: COOPERATE ';' {$$ = numbefore ($1,Cooperate());} ;

halt: HALT ';' {$$ = numbefore ($1,Halt());} ;

return: RETURN ';' {$$ = numbefore ($1,Return());} ;

await:
   AWAIT {parenth_list=1;} PARENTH_LIST ';'
   {parenth_list=0; $$ = numbefore ($1,Await($3));}
   ;

join:
   JOIN {parenth_list=1;} PARENTH_LIST ';'
   {parenth_list=0; $$ = numbefore ($1,Join($3));}
   ;

get_value:
   GET_VALUE {parenth_list=1;} PARENTH_LIST ';'
   {parenth_list=0; $$ =  numbefore ($1,Get_Value($3));}
   ;

if_interm1: {parenth_exp=1;$$ = (void*)new_state();} ;
if_interm2: {$$ = (void*)new_state();} ;

if:
   IF if_interm1 PARENTH_EXP {parenth_exp=0;}
   then.branch if_interm2 else.branch END
   {$$ = numbefore($1,If((int)$2,$3,$5,(int)$6,$7));}
   ;

while_interm: {parenth_exp=1;$$ = (void*)new_state();} ;

while:
   WHILE_TOK while_interm PARENTH_EXP {parenth_exp=0;}
   DO inst END {$$ = numbefore($1,While((int)$2,$3,$6));}
   ;

repeat_interm: {parenth_exp=1;$$ = (void*)InitRepeat();} ;

repeat:
   REPEAT repeat_interm PARENTH_EXP {parenth_exp=0;}
   DO inst END {$$ = numbefore($1,Repeat((int)$2,$3,$6));}
   ;

run:
   RUN IDENTIFIER {parenth_exp=1;} PARENTH_EXP {parenth_exp=0;} ';'
   {$$ = numbefore($1,Run($2,$4));}
   ;

/************************************************************/
unlink: UNLINK ';' {$$ = numbefore($1,Unlink());} ;

link:
   LINK {parenth_exp=1;} PARENTH_EXP {parenth_exp=0;} ';'
   {$$ = numbefore($1,Link($3));}
   ;

/************************************************************/
then.branch:
     /* Empty */ {$$ = Nothing();}
   | THEN inst {$$ = numbefore($1,$2);}
   ;

else.branch:
     /* Empty */ {$$ = Nothing();}
   | ELSE inst {$$ = numbefore($1,$2);}
   ;

%%
#include "parse.c"

/************** END OF GRAMMAR **************/
