#ifndef NULL
#define NULL 0
#endif

#define local(k) (_local_vars->k)

/******** return codes ****************/
#define OK         0 // like for pthread: 0 means OK
#define EBADMEM    1 // not enough memory
#define ETIMEOUT   2 // timeout expired
#define ENEXT      3 // no value generated during last intant
#define EBADLINK   4 // bad linking to a scheduler
#define EBADCREATE 5 // bad creation

/************ macros for automata **********************/
#define AUTOMATON(name)                      _AUTOMATON(name)
#define DEFINE_AUTOMATON(name)               _DEFINE_AUTOMATON(name)
#define BEGIN_AUTOMATON                      _BEGIN_AUTOMATON        
#define END_AUTOMATON                        _END_AUTOMATON

#define GOTO(num)                            _GOTO(num) 
#define GOTO_NEXT                            _GOTO_NEXT
#define IMMEDIATE(num)                       _IMMEDIATE(num)
#define RETURN                               _THE_END

#define SELF                                 _SELF
#define SET_LOCAL(data)                      _SET_LOCAL(data)
#define LOCAL                                _LOCAL
#define ARGS                                 _ARGS
#define RETURN_CODE                          _RETURN_CODE

#define STATE(num)                           _STATE(num)
#define STATE_AWAIT(num,event)               _STATE(num) _AWAIT(event)
#define STATE_AWAIT_N(num,event,n)           _STATE(num) _AWAIT_N(event,n)
#define STATE_GET_VALUE(num,event,n,result)  _STATE(num) _GET_VALUE(event,n,result)
#define STATE_STAY(num,n)                    _STATE(num) _STAY_INST(n)
#define STATE_JOIN(num,thread)               _STATE(num) _JOIN(thread)
#define STATE_JOIN_N(num,thread,n)           _STATE(num) _JOIN_N(thread,n)
#define STATE_SELECT(num,len,array,mask)     _STATE(num) _SELECT(len,array,mask)
#define STATE_SELECT_N(num,len,array,mask,n) _STATE(num) _SELECT_N(len,array,mask,n)

/**********************************************
  internal definitions: should not be used
***********************************************/
#define _AUTOMATON(x) void x (ft_thread_t)

#define _DEFINE_AUTOMATON(name)\
   void name (ft_thread_t _self)\

#define _BEGIN_AUTOMATON\
      while (1) {\
         switch (_GET_STATE())\
         {	    
      
#define _STATE(n)\
            case n: _SET_STATE(n);

#define _END_AUTOMATON\
            default: _THE_END\
         }\
      }

#define _GOTO_NEXT    {_SET_STATE(_GET_STATE()+1); _SET_STATUS(_DONE); return;} // bug corrected: 7/16/02
#define _GOTO(n)      {_SET_STATE(n); _SET_STATUS(_DONE); return;}
#define _IMMEDIATE(n) {_SET_STATE(n); break;}
#define _THE_END      {_SET_STATE(-1); _SET_STATUS(_TERMINATED); _terminate (_self); return;}

/****************************************************/
#define _SELF            _self
#define _SET_LOCAL(data) _set_thread_local (_self,(void*)data)
#define _LOCAL           _get_thread_local (_self)
#define _ARGS            _get_thread_args  (_self)
#define _RETURN_CODE     _get_thread_return_code (_self)

/****************************************************/
#define _READY      0 // ready to run
#define _BLOCKED    2 // waiting to be reexecuted during the same instant
#define _DONE       3 // nothing to be done for current instant
#define _TERMINATED 4 // nothing to be done in the future
#define _SUSPENDED  5 // suspended, waiting to be resumed
#define _WAIT       6 // waiting for an event to be generated
#define _STAY       7 // automaton stays in the same state

/****************************************************/
#define _SET_STATE(n)    _set_thread_autom_state  (_self,n)
#define _GET_STATE()     _get_thread_autom_state  (_self)
#define _SET_STATUS(s)   _set_thread_status       (_self,s)
// bug corrected 9/12/02
/****************************************************/
#define _DEADLINE         _get_thread_deadline   (_self)
#define _SET_DEADLINE(n)  _set_thread_deadline   (_self,n)
#define _CURRENT_INSTANT  _get_scheduler_instant (sched)

/****************************************************/
#define _STAY_OR_PASS(ret)\
  {int b = ret; if (b == _STAY) return; else _set_thread_return_code (_self,b);}

#define _STAY_INST(delay)               _STAY_OR_PASS(_automaton_stay      (_self,delay))
#define _AWAIT(event)                   _STAY_OR_PASS(_automaton_await     (_self,event))
#define _AWAIT_N(event,delay)           _STAY_OR_PASS(_automaton_await_n   (_self,event,delay))
#define _GET_VALUE(event,index,result)  _STAY_OR_PASS(_automaton_get_value (_self,event,index,result))
#define _JOIN(thread)                   _STAY_OR_PASS(_automaton_join      (_self,thread))
#define _JOIN_N(thread,delay)           _STAY_OR_PASS(_automaton_join_n    (_self,thread,delay))
#define _SELECT(len,array,mask)         _STAY_OR_PASS(_automaton_select    (_self,len,array,mask))
#define _SELECT_N(len,array,mask,delay) _STAY_OR_PASS(_automaton_select_n  (_self,len,array,mask,delay))

#define STATE_AWAIT_COND(num,cond)          _STATE(num) _AWAIT_COND(cond)
#define STATE_AWAIT_COND_N(num,cond,delay)  _STATE(num) _AWAIT_COND_N(cond,delay)

#define _AWAIT_COND(cond)                   _STAY_OR_PASS(_automaton_await_cond(_self,cond))
#define _AWAIT_COND_N(cond,delay)           _STAY_OR_PASS(_automaton_await_cond_n(_self,cond,delay))

/******************************************************/
